#ifndef _ARTSPLAYER_CPP_
#define _ARTSPLAYER_CPP_

#include <iostream>
#include <string>

using namespace std;


#include "config.h"

#include "artsPlayer.h"
#include "reference.h"
#include "part.h"
#include "track.h"
#include "note.h"
#include "midiEvent.h"
#include "masterEvent.h"
#include "audioEvent.h"
#include "song.h"
#include "table.h"
#include "element.h"
#include "scoreTrack.h"
#include "drumTrack.h"
#include "prProgress.h"
#include "prMainEditor.h"
#include "songIterator.h"
#include "addElement.h"

#include "artsmidi.h"
#include "soundserver.h"
#include "BrahmsArts.h"
#include "connect.h"

extern Table * selectioN;

using namespace Arts;

struct PlayerState {
  TimeStamp playTime;
  TimeStamp startTime;
  MidiManager midiManager;
  MidiPort midiPort;
  MidiClient midiClient;
  SoundServer soundServer;
  BrahmsMidiRecorder midiRecorder;
  MidiClient midiClientRec;
  Part * record_midi_part;
  Part * record_audio_part;
  int record_track;
  long pp;
  int ntracks;
  SongIterator * iterator;
  long left;
  long right;
  double interval;
  // unsigned long volume;
#ifdef HAVE_ARTS_ENVIRONMENT
  AudioSync audioSync;
  std::list<Synth_PLAY_WAV> audioList;
  Synth_CAPTURE_WAV captureWav;
#endif
};


extern PrMainEditor * mainEditor;


ArtsPlayer::ArtsPlayer() : Player() {
  init();
}

ArtsPlayer::ArtsPlayer(Song * s) : Player(s) {
  init();
}

ArtsPlayer::~ArtsPlayer() {
  freePlayer();
  // TODO: delete devman; // causes a segfault invalid filedescriptor
}



void ArtsPlayer::init() {
  _player_name = strdup("Arts");
  initPlayer();
}

const char * ArtsPlayer::device(int i) {
  const char * dev = 0;
  dev = "aRts";
  return dev;
}

int ArtsPlayer::devicenum() {
  int ret = 0;
  ret = 1;
  return ret;
}

void ArtsPlayer::play(PrProgress * progress, Part * part, long left, long right, bool cycl, int record_track) {
  ps->record_track = record_track;
  
  if (record_track!=-1) {
    Part * pt = (Part *)((::Reference*)selectioN->first())->getValue();
    if (pt->track()->isA()==AUDIOTRACK) {
      ps->record_audio_part = pt;
      ps->record_midi_part = 0;
    } else {
      ps->record_audio_part = 0;
      ps->record_midi_part = pt;
    }
  } else {
    ps->record_midi_part = 0;
    ps->record_audio_part = 0;
  }
  playStart(progress, part, left, right);
  mainEditor->startTimer();
}



void ArtsPlayer::hit(int out, int ch, int freq, int vol) {
  if (!_playing) {
    _playing = true;
    
    ps->playTime = ps->midiPort.time();
    // ps->midiPort.processCommand(MidiCommand(mcsProgram | ch, 0 /*=prg*/, 0));
    Arts::MidiEvent ev;
    ev.time = ps->playTime;
    ev.command = MidiCommand(mcsNoteOn|ch, freq, vol);
    ps->midiPort.processEvent(ev);
    
    ps->playTime.usec += 60*1000000/_song->tempo(); // 1/4 note
    ps->playTime.sec += ps->playTime.usec / 1000000;
    ps->playTime.usec %= 1000000;
    
    ev.time = ps->playTime;
    ev.command = MidiCommand(mcsNoteOff|ch, freq, 0);
    ps->midiPort.processEvent(ev);
    _playing = false;
  }
}

void ArtsPlayer::quad(int out, int ch, int freq1, int freq2, int freq3, int freq4, int vol) {
  if (!_playing) {
    _playing = true;

    cout << "TODO: ARTS: Player::quad" << endl;
    
    _playing = false;
  }
}





static float deltaT(TimeStamp a, TimeStamp b)
{
  return ((float)(a.sec - b.sec) + (float)(a.usec - b.usec)/1000000.0);
}


double ArtsPlayer::time() {
  return ((double)(ps->midiPort.time().sec - ps->startTime.sec) + (double)(ps->midiPort.time().usec - ps->startTime.usec)/1000000.0);
}
  
long ArtsPlayer::ticks() {
  return long(_song->tempo()*384*time()/60);
}

long ArtsPlayer::volume() {
  return 0; //ps->volume;
}

void ArtsPlayer::initPlayer() {
    ps = new PlayerState;
    ps->iterator = 0;
    bool veto = false;

    ps->soundServer = Arts::Reference("global:Arts_SoundServer");
    if (ps->soundServer.isNull()) {
      mainEditor->fatalError(PrMainEditor::errorArtsConnect);
      cout << "soundServer is null" << endl; veto = true;
    }

    ps->midiManager = Arts::Reference("global:Arts_MidiManager");
    if (ps->midiManager.isNull()) {
      mainEditor->fatalError(PrMainEditor::errorArtsMidi);
      cout << "midiManager is null" << endl; veto = true;
    }

    ps->midiClient = ps->midiManager.addClient(mcdPlay, mctApplication, "Brahms", "brahms");
    if (ps->midiClient.isNull()) { cout << "midiClient is null" << endl; veto = true; }

    ps->midiPort = ps->midiClient.addOutputPort();
    if (ps->midiPort.isNull()) { cout << "midiOutputPort is null" << endl; veto = true; }

    ps->midiRecorder = Arts::DynamicCast(ps->soundServer.createObject("BrahmsMidiRecorder"));
    if (ps->midiRecorder.isNull()) { 
      mainEditor->fatalError(PrMainEditor::errorArtsMidiRecord);
      cout << "midiRecorder is null" << endl; veto = true;
    }

    ps->midiClientRec = ps->midiManager.addClient(mcdRecord, mctApplication, "Brahms", "brahms");
    if (ps->midiClientRec.isNull()) { cout << "midiClientRec is null" << endl; veto = true; }

    ps->midiClientRec.addInputPort(ps->midiRecorder);
    ps->midiRecorder.setMidiThru(ps->midiPort);
}

void ArtsPlayer::freePlayer()
{
  delete ps;
}

bool ArtsPlayer::playStart(PrProgress * progress, Part * part, long lft, long rgt) {
  if (!_playing) {
    _playing = true;

    if (ps->record_midi_part!=0) ps->midiRecorder.record(true);
    // if (ps->record_audio_part!=0): SEE BELOW!!! ps->audioRecorder.record(true);
    ps->ntracks = _song->size();
    ps->left = lft;
    ps->right = rgt;
    ps->iterator = new SongIterator(_song, lft, rgt);
    ps->interval = TIMERINTERVALL * 0.001;
    bool veto = false;
    // ps->volume = 0;
    
    // TODO: handle track->output()
    
#ifdef HAVE_ARTS_ENVIRONMENT
    Track * tr = ps->iterator->track(0);
    if( tr )
      ps->midiPort = tr->envInstrument().port();
#endif
    
    
    ps->playTime = ps->midiPort.time(); // + xy in order to avoid neglecting the first events...
    //ps->playTime.usec += 200000;
    //ps->playTime.sec += ps->playTime.usec / 1000000;
    ps->startTime = ps->playTime; // remember start time
    
    if (!veto) {
      
      for (int t=0; t<ps->ntracks; t++) {
	if (ps->iterator->channel(t)!=-1) {
#ifdef HAVE_ARTS_ENVIRONMENT
	  ps->iterator->track(t)->envInstrument().port().processCommand(MidiCommand(mcsProgram | ps->iterator->channel(t), ps->iterator->program(t), 0));
#else
	  ps->midiPort.processCommand(MidiCommand(mcsProgram | ps->iterator->channel(t), ps->iterator->program(t), 0));
#endif
	}
      }

#ifdef HAVE_ARTS_ENVIRONMENT  
      ps->audioSync = DynamicCast(ps->soundServer.createObject("Arts::AudioSync"));
      assert(!ps->audioSync.isNull());
      if (ps->record_audio_part!=0) {
	ps->captureWav = DynamicCast(ps->soundServer.createObject("Arts::Synth_CAPTURE_WAV"));
	ps->captureWav.filename("test.wav");
	Synth_BUS_DOWNLINK sbd = DynamicCast(ps->soundServer.createObject("Arts::Synth_BUS_DOWNLINK"));
	sbd.busname("in_soundcard");
	connect(sbd,ps->captureWav);
	ps->captureWav._addChild(sbd,"sbd");
	ps->audioSync.queueStart(ps->captureWav);
	ps->audioSync.queueStart(sbd);
	ps->audioSync.executeAt(ps->playTime);
      }
#endif
      ps->pp = ps->left;
      
      
      /* this is to adjust the start-position into the past, in order to get a correct value for time(),
       * which is displayed in the gui!
       */
      long delta_usec = long(ps->pp*60*1000000.0/(384.0*_song->tempo()));
      long delta_sec  = long(delta_usec * 1.0 / 1000000.0);
      delta_usec %= 1000000;
      ps->startTime.sec -= delta_sec;
      ps->startTime.usec -= delta_usec;
      if (ps->startTime.usec<0) {
	ps->startTime.usec += 1000000;
	ps->startTime.sec -= 1;
      }
      
    }
    // we are still playing! (was: _playing = false;)
  }
  return true;
}

bool ArtsPlayer::playQuantum(bool cycl) {
  bool ret = true;
    TimeStamp midiTime = ps->midiPort.time();
    Arts::MidiEvent ev;
    Event * event;
    Note * note;
    Track * tr;
    int chan;
    long num;
    ::MidiEvent * mev;

    while((deltaT(ps->playTime,midiTime) < ps->interval)
       && (ps->pp < ps->right || ps->right == 0)) {
      ev.time = ps->playTime;

      if (ps->record_midi_part!=0) metronome(ps->pp);

      if (!ps->iterator->done()) {
	for (int t=0; t<ps->ntracks; t++) {
	  tr = ps->iterator->track(t);
	  while (ps->iterator->startsAt(t,ps->pp)) {
	    if (!ps->iterator->mute(t)) {
	      event = (Event*) (*ps->iterator)[t];
	      if (tr->delay()!=0) {
		ev.time.usec += tr->delay()*1000000/(384*_song->tempo()); // 60000;   // 60ms
		ev.time.sec += ev.time.usec / 1000000;
		ev.time.usec %= 1000000;
	      }
	      if (event->isA()==NOTE) {
		note = (Note*) event;
		// cout << ps->pp << " on " << note->pitch() << endl;
		ev.command = MidiCommand(mcsNoteOn|ps->iterator->channel(t), tr->pitch(note), tr->vel(note));
#ifdef HAVE_ARTS_ENVIRONMENT
		tr->envInstrument().port().processEvent(ev);
#else
		ps->midiPort.processEvent(ev);
#endif
		ps->iterator->activate(t);
		// ps->volume = ps->volume|(1<<t);
		// cout << "* " << ps->volume << endl;
	      } else if (event->isA()==MASTEREVENT) {
		// TODO!
	      } else if (event->isA()==MIDIEVENT) {
		mev = (::MidiEvent*) event;
		if(mev->code() >= 10 && mev->code() <= 14) {
		  ev.command = MidiCommand((mev->code() << 4) | mev->channel(), mev->value1(), mev->value2());
#ifdef HAVE_ARTS_ENVIRONMENT
		tr->envInstrument().port().processEvent(ev);
#else
		  ps->midiPort.processEvent(ev);
#endif
		}
	      } else if (event->isA()==AUDIOEVENT) {
		cout << "play " << ((AudioEvent*)event)->fileName() << endl;
#ifdef HAVE_ARTS_ENVIRONMENT
		Synth_PLAY_WAV spw = DynamicCast(ps->soundServer.createObject("Arts::Synth_PLAY_WAV"));
		spw.filename(((AudioEvent*)event)->fileName());
		ps->audioList.push_back(spw);
		Synth_BUS_UPLINK sbu = DynamicCast(ps->soundServer.createObject("Arts::Synth_BUS_UPLINK"));
		sbu.busname("out_soundcard");
		connect(spw,sbu);
		spw._addChild(sbu,"sbu");
		ps->audioSync.queueStart(spw);
		ps->audioSync.queueStart(sbu);
		ps->audioSync.executeAt(ev.time);
#endif
	      } else {
		cout << "track " << t << *event << endl;
	      }
	      if (tr->delay()!=0) {
		ev.time = ps->playTime;
	      }
	    }
	    ps->iterator->increment(t); // TODO: new program settings?
	  }
	}
      }
#ifdef HAVE_ARTS_ENVIRONMENT
      if (!ps->audioList.empty()) tidyAudioList();
#endif
      ps->pp++;

      ps->playTime.usec += 60*1000000/(384*_song->tempo()); // 60000;   // 60ms
      ps->playTime.sec += ps->playTime.usec / 1000000;
      ps->playTime.usec %= 1000000;

      if (!ps->iterator->done()) {
	while (ps->iterator->activeEndsAt(ps->pp)) {
	  note = ps->iterator->activeNote();
	  chan = ps->iterator->activeChannel();
	  num  = ps->iterator->activeTrack();
	  tr   = ps->iterator->track(num);
	  // cout << ps->pp << " off " << note->pitch() << endl;
	  ev.command = MidiCommand(mcsNoteOff|chan, tr->pitch(note), tr->vel(note));
#ifdef HAVE_ARTS_ENVIRONMENT
		tr->envInstrument().port().processEvent(ev);
#else
	  ps->midiPort.processEvent(ev);
#endif
	  ps->iterator->cutActive();
	  // ps->volume = ps->volume&((-1)-(1<<num));
	  // cout << "- " << ps->volume << endl;
	}
      }

      if (cycl && ps->iterator->done() && ps->pp == ps->right) {
	// looping over the end of an event will not work correctly, i.e.
	// if your song is
	//     left           right
	//
	//      |[---]          |
	//      |-]   [---]   [-|-]
	//      |     [---]   [-|-]
	//
	// then the events that cross the left/right mark should be stopped/
	// restarted - we don't do this yet

	ps->iterator = new SongIterator(_song, ps->left, ps->right);
	ps->pp = ps->left;
	ps->startTime = ps->playTime; // remember start time
	// TODO: new program settings
      }
    }


    /*
     * to give the caller a realistic idea when we are "really done", i.e. when
     * no more events are in the play queue, check that the play queue is empty
     * before saying "there is nothing more to play"
     */
    // bool reallyDone=(ps->iterator->done() && deltaT(ps->playTime, midiTime) < 0);
    bool reallyDone=(ps->pp>=ps->right && deltaT(ps->playTime, midiTime) < 0);
    ret= !reallyDone;
  return ret;
}

void ArtsPlayer::tidyAudioList() {
#ifdef HAVE_ARTS_ENVIRONMENT
  list<Synth_PLAY_WAV>::iterator i = ps->audioList.begin();
  while (i!=ps->audioList.end()) {
    if (i->finished()) { i = ps->audioList.erase(i); cout << "tidyAudioList::erase " << endl; }
    else i++;
  }
#endif
}


void ArtsPlayer::killAllNotes() {
    ps->midiRecorder.record(false);
    Arts::MidiEvent ev;
    ev.time = ps->playTime;
    for (int ch=0; ch<16; ch++) {
      // TODO: after KDE2.2 has been released, switch to
      // MidiCommand(mcsParameter|ch, mcpAllNotesOff, 0);
      // but as this will not compile for KDE2.1 users, hardcode it:
      ev.command = MidiCommand(mcsParameter|ch, 0x7b, 0);
#ifdef HAVE_ARTS_ENVIRONMENT
      if (ps->iterator!=0)
	for (int t=0; t<ps->ntracks; t++)
	  ps->iterator->track(t)->envInstrument().port().processEvent(ev);
#else
      ps->midiPort.processEvent(ev);
#endif
    }
    if (ps->iterator!=0) delete ps->iterator;
    ps->iterator = 0;

    mergeQueue();
#ifdef HAVE_ARTS_ENVIRONMENT
    ps->audioList.clear();
    ps->audioSync = AudioSync::null();
    ps->captureWav = Synth_CAPTURE_WAV::null();
#endif

  _playing = false;
}

void ArtsPlayer::setTempo(int t, long pos) {
  // TODO!
}

void ArtsPlayer::setPosition(Position p) {
  ps->pp = p.ticks();
}





#define MAXSUS 512


void ArtsPlayer::mergeQueue() {
  if (ps->record_midi_part!=0) {
    std::vector<Arts::MidiEvent> * q = ps->midiRecorder.getQueue();
    cout << "merge " << endl;
    Note* sus[MAXSUS];
    Note* newNote;
    int susp=0;

    int remaining = q->size();
    /* for (std::vector<Arts::MidiEvent>::iterator i=q->begin(); i!=q->end(); i++) {
       cout << i->time.sec << ":" << i->time.usec << ", " << i->command.status << endl;
       }*/
    std::vector<Arts::MidiEvent>::iterator qiter = q->begin();

    if (ps->iterator!=0) delete ps->iterator;
    ps->iterator = new SongIterator(_song, ps->left, ps->right);
    ps->pp = ps->left;
    ps->playTime = ps->startTime;
    int ch = 0;
    int cmd = 0;
    int ii = 0;
    MidiCommand mc;

    for (; ps->pp < ps->right && remaining!=0;) {
      if (deltaT(qiter->time, ps->playTime) <=0) {
	mc = qiter->command;
	ch = mc.status&0x0f;
	cmd = mc.status&0xf0;
	if (cmd==0x90 && mc.data2==0) cmd=0x80;
	// cout << "add: " << ps->pp << ", cmd: " << cmd << ", ch: " << ch << ", note:" << int(mc.data1) << ", " << int(mc.data2) << endl;
	switch (cmd) {
	case 0x80: // NOTEOFF
	  ii=0;
	  for (int i=0; i<susp; i++)
	    if ((mc.data1 == sus[i]->pitch()))
	      {
		sus[i]->setDuration(ps->pp - ps->record_midi_part->start(sus[i]).ticks());
		ii=i;
		i=susp;
	      }
	  for (int i=ii; i<susp-1; i++) sus[i]=sus[i+1];
	  if (susp>0) susp--;
	  break;
	case 0x90: // NOTEON
	  newNote = new Note(mc.data1, mc.data2, 191, ps->pp-ps->record_midi_part->start(), 0, ch);
	  // ps->record_midi_part->add(newNote);
	  _song->doo(new AddElement(newNote,ps->record_midi_part));
	  if (susp<MAXSUS) sus[susp++] = newNote;
	  break;
	}

	qiter++;
	remaining--;
      }

      // TODO: cycles!
      ps->pp++;
      ps->playTime.usec += 60*1000000/(384*_song->tempo()); // 60000;   // 60ms
      ps->playTime.sec += ps->playTime.usec / 1000000;
      ps->playTime.usec %= 1000000;
    }
  }
  if (ps->iterator!=0) delete ps->iterator;
  ps->iterator = 0;
}


void ArtsPlayer::programChange(Track * tr) {
  int ch  = tr->channel();
  int prg = tr->program();
  ps->midiPort.processCommand(MidiCommand(mcsProgram | ch, prg, 0));
}


void ArtsPlayer::metronome(int p) {
  if (p%384==0) {
    Arts::MidiEvent ev;
    ev.time = ps->playTime;
    ev.command = MidiCommand(mcsNoteOn|met_chn, met_pitch, met_vel);
    ps->midiPort.processEvent(ev);
    ev.time = ps->playTime;
    ev.time.usec += 6000;
    ev.command = MidiCommand(mcsNoteOff|met_chn, met_pitch, met_vel);
    ps->midiPort.processEvent(ev);
  }
}

// vim:ts=8:sw=2:sts=2

#endif
