#ifndef _PLAYER_H_
#define _PLAYER_H_

#include "position.h"

class Song;
class Part;
class PrProgress;
class SongIterator;

enum { NO_PLAYER, ARTS_PLAYER, ALSA_5_PLAYER, ALSA_9_PLAYER };

#define MAXDEVNUM 16
#define TIMERINTERVALL 1000.0

class Player
{
  friend class Track;

 protected:
  Song * _song;
  int devnum;
  bool _playing;
  char * devices[MAXDEVNUM];
  long oldTimePos;
  bool needSync;
  int ntracks;
  long pp;
  long left;
  int met_chn;
  int met_pitch;
  int met_vel;
  const char * _player_name;

  virtual void init()=0;
  // virtual void wait(long,long) = 0;
  virtual void mergeQueue() = 0;
  virtual void metronome(int) = 0;
  virtual void initPlayer() = 0;
  virtual void freePlayer() = 0;

 public:
  Player();
  Player(Song*);
  virtual ~Player();
  virtual const char * device(int i) = 0;
  virtual int devicenum() = 0;
  virtual void play(PrProgress * progress = 0, Part * part=0, long left = 0, long right = 0, bool cycl = false, int record_track = -1) = 0;
  virtual void hit(int out, int ch, int freq, int vol) = 0;
  virtual void quad(int out, int ch, int freq1, int freq2, int freq3, int freq4, int vol) = 0;
  virtual void killAllNotes() = 0;
  virtual void setTempo(int t, long pos=0) = 0;
  virtual bool playQuantum(bool cycl = false) = 0;
  virtual bool playStart(PrProgress * progress, Part * part, long left, long right) = 0;
  virtual double time()= 0;
  virtual long ticks() = 0;
  virtual long volume() = 0;
  virtual void programChange(Track*) = 0;
  virtual void setPosition(Position p) = 0;

  void programChange(int track);
  bool playing() const { return _playing; }
  const char * name() { return _player_name; }
};

#endif
