/***************************************************************************
                          sequence.h  -  description
                             -------------------
    begin                : Sun Jun 9 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SEQUENCE_H
#define SEQUENCE_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// include files for Qt
#include <qwidget.h>

/**
  *@author Morten Hustveit
  */

class QPainter;
class QScrollBar;
class QLabel;
class QGridLayout;
class KPopupMenu;
class SanetrackerDoc;
class Scheduler;

class Sequence : public QWidget
{
  Q_OBJECT

public:

  Sequence(QWidget* parent, SanetrackerDoc* doc, const Scheduler& scheduler);
  ~Sequence();

  void reconfigure();

  uint getCurrentRow() { return this->currentRow; }
  uint getCurrentCol() { return this->currentCol; }

  bool active() const { return _active; }
  void setActive(bool _active = true);

public slots:

  void keyPressed(QKeyEvent*);
  void insertColumn();
  void removeColumn();
  void clearRow();
  void positionChanged();

signals:

  void activated();
  void selectionChanged(uint device, uint pattern, uint position);

protected:

  void resizeEvent(QResizeEvent*);
  void wheelEvent(QWheelEvent*);

private:

  SanetrackerDoc* doc;

  bool _active;

  friend class Labels : public QWidget
  {
  public:

    Labels(Sequence* parent);

    Sequence* parent;

    void paintRow(QPainter& painter, uint row);
    void paintEvent(QPaintEvent*);
    void mousePressEvent(QMouseEvent*);
  };

  friend class PosLabels : public QWidget
  {
  public:

    PosLabels(Sequence* parent);

    Sequence* parent;

    void paintEvent(QPaintEvent*);
  };

  friend class Data : public QWidget
  {
  public:

    Data(Sequence* parent);

    Sequence* parent;

    void paintCell(QPainter& painter, uint row, uint col);
    void paintEvent(QPaintEvent*);
    void mousePressEvent(QMouseEvent*);
  };

  uint         labelWidth;
  uint         labelHeight;
  uint         currentRow;
  uint         currentCol;

  KPopupMenu*  menu;
  QGridLayout* layout;
  QScrollBar*  scrollVert;
  QScrollBar*  scrollHoriz;
  Labels*      labels;
  PosLabels*   posLabels;
  Data*        data;

  const Scheduler& scheduler;

protected slots:

  void slotHorizScroll(int val);
  void slotHorizReleased();
  void slotVertScroll(int val);
  void slotNextPattern();
  void slotPrevPattern();
};

#endif
